@@ -186,8 +186,8 @@ void cleanup_client(struct client_s *client)
 {
     assuan_context_t ctx = client->ctx;
     struct client_thread_s *thd = client->thd;
-    gboolean lock_on_open = client->lockonopen;
     gboolean rc_on_locked = client->rc_on_locked;
+    gboolean lock_on_open = client->lock_on_open;
 #ifdef WITH_PINENTRY
     struct pinentry_s *pin = client->pinentry;
 #endif
@@ -214,7 +214,7 @@ void cleanup_client(struct client_s *client)
     client->pinentry = pin;
 #endif
     client->rc_on_locked = rc_on_locked;
-    client->lockonopen = lock_on_open;
+    client->lock_on_open = lock_on_open;
 }
 
 static void gz_cleanup(void *arg)
@@ -488,7 +488,7 @@ done:
 
     cleanup_crypto(&client->crypto);
 
-    if (!rc && client->lockonopen)
+    if (!rc && client->lock_on_open)
 	return do_lock_command(client);
 
     return send_error(ctx, rc);
@@ -502,6 +502,52 @@ static void req_cleanup(void *arg)
     g_strfreev((gchar **)arg);
 }
 
+static gpg_error_t parse_open_opt_lock(gpointer data, gpointer value)
+{
+    struct client_s *client = data;
+    const gchar *p = value;
+
+    if (p && *p) {
+	long l = strtol(p, NULL, 10);
+
+	if (l < 0 || l > 1)
+	    return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
+
+	client->lock_on_open = l ? TRUE : FALSE;
+    }
+    else if ((!p || !*p) && (client->opts & OPT_LOCK))
+	client->lock_on_open = FALSE;
+    else
+	client->lock_on_open = TRUE;
+
+    return 0;
+}
+
+static gpg_error_t parse_opt_pinentry(gpointer data, gpointer value)
+{
+#ifdef WITH_PINENTRY
+    struct client_s *client = data;
+    gchar *p = NULL;
+    gchar *str = value;
+    gint n;
+
+    if (!str || !*str) {
+	client->pinentry->enable = -1;
+	client->opts &= ~(OPT_PINENTRY);
+	return 0;
+    }
+
+    n = strtol(str, &p, 10);
+
+    if (*p || n < 0 || n > 1)
+	return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
+
+    client->pinentry->enable = n ? TRUE : FALSE;
+    client->opts |= OPT_PINENTRY;
+#endif
+    return 0;
+}
+
 static gint open_command(assuan_context_t ctx, gchar *line)
 {
     gboolean cached = FALSE;
@@ -510,6 +556,22 @@ static gint open_command(assuan_context_t ctx, gchar *line)
     gchar **req;
     gchar *filename = NULL;
     gsize hashlen = gcry_md_get_algo_dlen(GCRY_MD_SHA256);
+    struct argv_s *args[] = {
+	&(struct argv_s) { "lock", OPT_NOARG, parse_open_opt_lock },
+	&(struct argv_s) { "pinentry", OPT_OPTARG, parse_opt_pinentry},
+	NULL
+    };
+
+    if (client->state == STATE_OPEN)
+	cleanup_client(client);
+
+    if (!(client->opts & OPT_LOCK))
+	client->lock_on_open = FALSE;
+
+    rc = parse_options(&line, args, client);
+
+    if (rc)
+	return send_error(ctx, rc);
 
     if ((req = split_input_line(line, " ", 2)) != NULL)
 	filename = req[0];
@@ -528,9 +590,6 @@ static gint open_command(assuan_context_t ctx, gchar *line)
 	return send_error(ctx, GPG_ERR_INV_VALUE);
     }
 
-    if (client->state == STATE_OPEN)
-	cleanup_client(client);
-
     gcry_md_hash_buffer(GCRY_MD_MD5, client->md5file, filename, strlen(filename));
     CACHE_LOCK(client->ctx);
 
@@ -618,8 +677,13 @@ static gint open_command(assuan_context_t ctx, gchar *line)
 #endif
 	return open_command_finalize(ctx, NULL, cached);
     }
-    else
+    else {
+	if (!(client->opts & OPT_CIPHER))
+	    g_key_file_set_string(keyfileh, filename, "cipher",
+		    pwmd_cipher_to_str(client->crypto->fh->ver.fh2.flags));
+
 	client->mtime = client->crypto->fh->st.st_mtime;
+    }
 
     client->filename = g_strdup(filename);
 
@@ -1277,12 +1341,92 @@ static gpg_error_t save_command_finalize(assuan_context_t ctx, guchar *key,
     return send_error(ctx, 0);
 }
 
+static gpg_error_t parse_save_opt_iterations(gpointer data, gpointer v)
+{
+    struct client_s *client = data;
+    guint64 n;
+    gchar *value = v;
+    gchar *p = NULL;
+
+    if (!client->filename)
+	return EPWMD_NO_FILE;
+
+    if (!value || !*value)
+	return 0;
+
+    errno = 0;
+    n = strtoul(value, &p, 10);
+
+    if (errno || (p && *p) || n == G_MAXULONG)
+	return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
+
+    MUTEX_LOCK(&rcfile_mutex);
+    g_key_file_set_double(keyfileh,
+	    client->filename ? client->filename : "global", "iterations", n);
+    MUTEX_UNLOCK(&rcfile_mutex);
+
+    if (client->filename)
+	client->opts |= OPT_ITERATIONS;
+
+    return 0;
+}
+
+static gpg_error_t parse_save_opt_cipher(gpointer data, gpointer value)
+{
+    struct client_s *client = data;
+    const gchar *p = value;
+    guint64 flags;
+
+    if (!client->filename)
+	return EPWMD_NO_FILE;
+
+    if (!p || !*p)
+	return 0;
+
+    flags = pwmd_cipher_str_to_cipher(p);
+
+    if (!flags)
+	return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
+
+    MUTEX_LOCK(&rcfile_mutex);
+    g_key_file_set_string(keyfileh, client->filename, "cipher", p);
+    MUTEX_UNLOCK(&rcfile_mutex);
+
+    if (!value)
+	g_free((gchar *)p);
+
+    return 0;
+}
+
+static gpg_error_t parse_save_opt_reset(gpointer data, gpointer value)
+{
+    struct client_s *client = data;
+
+    CACHE_LOCK(client->ctx);
+    cache_clear(client->md5file, 1);
+    CACHE_UNLOCK;
+    return 0;
+}
+
 static gint save_command(assuan_context_t ctx, gchar *line)
 {
     gboolean cached = FALSE;
     struct stat st;
     struct client_s *client = assuan_get_pointer(ctx);
     guint hashlen = gcry_md_get_algo_dlen(GCRY_MD_SHA256);
+    gpg_error_t rc;
+    struct argv_s *args[] = {
+	&(struct argv_s) { "iterations", OPT_OPTARG, parse_save_opt_iterations },
+	&(struct argv_s) { "cipher", OPT_OPTARG, parse_save_opt_cipher },
+	&(struct argv_s) { "pinentry", OPT_OPTARG, parse_opt_pinentry},
+	&(struct argv_s) { "reset", OPT_NOARG, parse_save_opt_reset},
+	NULL
+    };
+
+    rc = parse_options(&line, args, client);
+
+    if (rc)
+	return send_error(ctx, rc);
 
     if (line && *line)
 	log_write2("ARGS=%s", "<passphrase>");
@@ -1442,19 +1586,12 @@ static gint delete_command(assuan_context_t ctx, gchar *line)
      * No sub-node defined. Remove the entire node (root element).
      */
     if (!req[1]) {
-	xmlNodePtr parent = n->parent;
-
 	if (n) {
-	    xmlUnlinkNode(n);
+	    rc = unlink_node(n);
 	    xmlFreeNode(n);
 	}
 
 	g_strfreev(req);
-	rc = 0;
-
-	if (parent)
-	    rc = update_element_mtime(parent);
-
 	return send_error(ctx, rc);
     }
 
@@ -1465,14 +1602,8 @@ static gint delete_command(assuan_context_t ctx, gchar *line)
 	return send_error(ctx, rc);
 
     if (n) {
-	xmlNodePtr parent = n->parent;
-
-	xmlUnlinkNode(n);
+	rc = unlink_node(n);
 	xmlFreeNode(n);
-	rc = 0;
-
-	if (parent)
-	    rc = update_element_mtime(parent);
     }
 
     return send_error(ctx, rc);
@@ -1852,59 +1983,96 @@ static void list_command_cleanup2(void *arg)
     struct element_list_s *elements = arg;
 
     if (elements) {
-	gint total = g_slist_length(elements->list);
-	gint i;
+	if (elements->list) {
+	    gint total = g_slist_length(elements->list);
+	    gint i;
 
-	for (i = 0; i < total; i++) {
-	    gchar *tmp = g_slist_nth_data(elements->list, i);
-	    g_free(tmp);
+	    for (i = 0; i < total; i++) {
+		gchar *tmp = g_slist_nth_data(elements->list, i);
+		g_free(tmp);
+	    }
+
+	    g_slist_free(elements->list);
 	}
 
-	g_slist_free(elements->list);
-
 	if (elements->prefix)
 	    g_free(elements->prefix);
 
+	if (elements->req)
+	    g_strfreev(elements->req);
+
 	g_free(elements);
     }
 }
 
+static gpg_error_t parse_list_opt_norecurse(gpointer data, gpointer value)
+{
+    struct element_list_s *elements = data;
+
+    elements->recurse = FALSE;
+    return 0;
+}
+
+static gpg_error_t parse_list_opt_verbose(gpointer data, gpointer value)
+{
+    struct element_list_s *elements = data;
+
+    elements->verbose = TRUE;
+    return 0;
+}
+
 static gint list_command(assuan_context_t ctx, gchar *line)
 {
     struct client_s *client = assuan_get_pointer(ctx);
     gpg_error_t rc;
     struct element_list_s *elements = NULL;
     gchar *tmp;
+    struct argv_s *args[] = {
+	&(struct argv_s) { "no-recurse", OPT_NOARG, parse_list_opt_norecurse },
+	&(struct argv_s) { "verbose", OPT_NOARG, parse_list_opt_verbose },
+	NULL
+    };
 
     if (disable_list_and_dump == TRUE)
 	return send_error(ctx, GPG_ERR_NOT_IMPLEMENTED);
 
-    log_write2("ARGS=\"%s\"", line);
-
-    if (!*line) {
-	GString *str;
-
-	rc = list_root_elements(client->doc, &str);
-
-	if (rc)
-	    return send_error(ctx, rc);
-
-	pth_cleanup_push(list_command_cleanup1, str);
-	rc = xfer_data(ctx, str->str, str->len);
-	pth_cleanup_pop(1);
-	return send_error(ctx, rc);
-    }
-
     elements = g_malloc0(sizeof(struct element_list_s));
 
     if (!elements) {
 	log_write("%s(%i): %s", __FILE__, __LINE__, strerror(ENOMEM));
-	rc = gpg_err_code_from_errno(ENOMEM);
-	goto fail;
+	return gpg_err_code_from_errno(ENOMEM);
     }
 
+    elements->recurse = TRUE; // default
     pth_cleanup_push(list_command_cleanup2, elements);
-    rc = create_path_list(client->doc, elements, line);
+    rc = parse_options(&line, args, elements);
+
+    if (rc)
+	goto fail;
+
+    if (!*line) {
+	GString *str;
+
+	rc = list_root_elements(client->doc, &str, elements->verbose);
+
+	if (rc) {
+	    pth_cleanup_pop(1);
+	    return send_error(ctx, rc);
+	}
+
+	pth_cleanup_push(list_command_cleanup1, str);
+	rc = xfer_data(ctx, str->str, str->len);
+	pth_cleanup_pop(1);
+	pth_cleanup_pop(1);
+	return send_error(ctx, rc);
+    }
+
+    elements->req = split_input_line(line, " ", 0);
+
+    if (!elements->req)
+	strv_printf(&elements->req, "%s", line);
+
+    rc = create_path_list(client->doc, elements, *elements->req);
 
     if (rc)
 	goto fail;
@@ -2668,11 +2836,7 @@ static gint getconfig_command(assuan_context_t ctx, gchar *line)
     }
     else if (!g_ascii_strcasecmp(paramp, "pinentry_timeout")) {
 #ifdef WITH_PINENTRY
-	if (fp == client->filename && (client->opts & OPT_PINENTRY_TO))
-	    p = g_strdup_printf("%i", client->pinentry->timeout);
-	else
-	    p = g_strdup_printf("%i",
-		    get_key_file_integer(fp, "pinentry_timeout"));
+	p = g_strdup_printf("%i", get_key_file_integer(fp, "pinentry_timeout"));
 
 	if (!p) {
 	    log_write("%s(%i): %s", __FILE__, __LINE__,
@@ -3028,6 +3192,9 @@ static gint import_command_finalize(gpointer data, gint assuan_rc, guchar *line,
 	n = xmlAddChildList(xmlDocGetRootElement(client->doc), n);
     }
 
+    if (n && !rc)
+	rc = update_element_mtime(n->parent);
+
 fail:
     if (doc)
 	xmlFreeDoc(doc);
@@ -3118,6 +3285,7 @@ static gint set_unset_common(assuan_context_t ctx, const gchar *name,
 	const gchar *value)
 {
     struct client_s *client = assuan_get_pointer(ctx);
+    gpg_error_t rc;
 
     if (g_ascii_strcasecmp(name, (gchar *)"log_level") == 0) {
 	glong l = 0;
@@ -3129,7 +3297,6 @@ static gint set_unset_common(assuan_context_t ctx, const gchar *name,
 		return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
 	}
 
-	log_write1("log_level=%li", l);
 	MUTEX_LOCK(&rcfile_mutex);
 	g_key_file_set_integer(keyfileh, "global", "log_level", (gint)l);
 	MUTEX_UNLOCK(&rcfile_mutex);
@@ -3145,82 +3312,37 @@ static gint set_unset_common(assuan_context_t ctx, const gchar *name,
 		return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
 	}
 
-	log_write1("rc_on_locked=%li", l);
 	client->rc_on_locked = l ? TRUE : FALSE;
 	goto done;
     }
     else if (g_ascii_strcasecmp(name, (gchar *)"lock_on_open") == 0) {
-	glong l = 0;
+	rc = parse_open_opt_lock(client, (gpointer)value);
 
-	if (value) {
-	    l = strtol(value, NULL, 10);
+	if (rc)
+	    return rc;
 
-	    if (l < 0 || l > 1)
-		return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
-	}
-
-	log_write1("lock_on_open=%li", l);
-	client->lockonopen = l ? TRUE : FALSE;
-	goto done;
+	client->opts |= OPT_LOCK;
     }
     else if (g_ascii_strcasecmp(name, (gchar *)"cipher") == 0) {
-	guint64 flags;
-	const gchar *p = value;
-
-	if (!client->filename)
-	    return EPWMD_NO_FILE;
-
-	if (value) {
-	    flags = pwmd_cipher_str_to_cipher(value);
-
-	    if (!flags)
-		return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
-	}
-	else if (!value)
-	    p = get_key_file_string("global", "cipher");
-
-	MUTEX_LOCK(&rcfile_mutex);
-	g_key_file_set_string(keyfileh, client->filename, "cipher", p);
-	MUTEX_UNLOCK(&rcfile_mutex);
-	log_write1("cipher=%s", p);
-
-	if (!value)
-	    g_free((gchar *)p);
-
-	goto done;
-    }
-    else if (g_ascii_strcasecmp(name, (gchar *)"iterations") == 0) {
-	guint64 n;
-	gchar *p = NULL;
-
-	if (!client->filename)
-	    return EPWMD_NO_FILE;
-
 	if (!value) {
-	    MUTEX_LOCK(&rcfile_mutex);
-	    g_key_file_set_double(keyfileh, client->filename, "iterations",
-		    get_key_file_double("global", "iterations"));
-	    MUTEX_UNLOCK(&rcfile_mutex);
-	    log_write1("iterations=%lu",
-		    (guint64)get_key_file_double(client->filename, "iterations"));
+	    client->opts &= ~(OPT_CIPHER);
 	    goto done;
 	}
 
-	errno = 0;
-	n = strtoul(value, &p, 10);
+	rc = parse_save_opt_cipher(client, (gpointer)value);
 
-	if (errno || (p && *p) || n == G_MAXULONG)
-	    return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
+	if (rc)
+	    return rc;
 
-	MUTEX_LOCK(&rcfile_mutex);
-	g_key_file_set_double(keyfileh,
-		client->filename ? client->filename : "global", "iterations", n);
-	MUTEX_UNLOCK(&rcfile_mutex);
+	client->opts |= OPT_CIPHER;
+	goto done;
+    }
+    else if (g_ascii_strcasecmp(name, (gchar *)"iterations") == 0) {
+	rc = parse_save_opt_iterations(client, (gpointer)value);
 
-	if (client->filename)
-	    client->opts |= OPT_ITERATIONS;
+	if (rc)
+	    return rc;
 
-	log_write1("iterations=%lu", n);
 	goto done;
     }
     else if (g_ascii_strcasecmp(name, (gchar *)"NAME") == 0) {
@@ -3235,7 +3357,6 @@ static gint set_unset_common(assuan_context_t ctx, const gchar *name,
 	print_fmt(buf, sizeof(buf), "%s", value);
 	pth_attr_set(attr, PTH_ATTR_NAME, buf);
 	pth_attr_destroy(attr);
-	log_write1("name=%s", buf);
 #ifdef WITH_PINENTRY
 	if (client->pinentry->name)
 	    g_free(client->pinentry->name);
@@ -3271,51 +3392,32 @@ static gint set_unset_common(assuan_context_t ctx, const gchar *name,
 	gchar *p = NULL;
 	gint n;
 	
-	if (!value) {
-	    client->pinentry->timeout =
-		get_key_file_integer(client->filename, "pinentry_timeout");
-	    client->opts &= ~(OPT_PINENTRY_TO);
-	    log_write1("pinentry_timeout=%i",
-		    get_key_file_integer(client->filename, "pinentry_timeout"));
+	if (!value)
 	    goto done;
-	}
 
 	n = strtol(value, &p, 10);
 
 	if (*p || n < 0)
 	    return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
 
-	client->pinentry->timeout = n;
-	client->opts |= OPT_PINENTRY_TO;
-	log_write1("pinentry_timeout=%i", n);
+	MUTEX_LOCK(&rcfile_mutex);
+	g_key_file_set_integer(keyfileh, client->filename ? client->filename :
+		"global", "pinentry_timeout", n);
+	MUTEX_UNLOCK(&rcfile_mutex);
 	goto done;
     }
     else if (g_ascii_strcasecmp(name, "enable_pinentry") == 0) {
-	gchar *p = NULL;
-	gint n;
-	
-	if (!value) {
-	    client->pinentry->enable = -1;
-	    client->opts &= ~(OPT_PINENTRY);
-	    goto done;
-	}
+	rc = parse_opt_pinentry(client, (gpointer)value);
 
-	n = strtol(value, &p, 10);
+	if (rc)
+	    return rc;
 
-	if (*p || n < 0 || n > 1)
-	    return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_INV_VALUE);
-
-	client->pinentry->enable = n == 0 ? FALSE : TRUE;
-	client->opts |= OPT_PINENTRY;
-	log_write1("enable_pinentry=%i", n);
 	goto done;
     }
 #endif
     else
 	return gpg_err_make(PWMD_ERR_SOURCE, GPG_ERR_UNKNOWN_OPTION);
 
-    log_write1("%s=%s", name, value ? value : "");
-
 done:
     return 0;
 }
@@ -3449,7 +3551,7 @@ static gint rename_command(assuan_context_t ctx, gchar *line)
     }
 
     if (ndst) {
-	xmlUnlinkNode(ndst);
+	unlink_node(ndst);
 	xmlFreeNodeList(ndst);
     }
 
@@ -3532,7 +3634,7 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 
     /* Merge any attributes from the src node to the initial dst node. */
     for (xmlAttrPtr attr = new->properties; attr; attr = attr->next) {
-	if ((!src[1] || !dst[1]) && xmlStrEqual(attr->name, (xmlChar *)"_name"))
+	if (xmlStrEqual(attr->name, (xmlChar *)"_name"))
 	    continue;
 
 	xmlAttrPtr a = xmlHasProp(ndst, attr->name);
@@ -3540,7 +3642,10 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 	if (a)
 	    xmlRemoveProp(a);
 
-	xmlNewProp(ndst, attr->name, attr->children->content);
+	xmlChar *tmp = xmlNodeGetContent(attr->children);
+	xmlNewProp(ndst, attr->name, tmp);
+	xmlFree(tmp);
+	rc = add_attribute(ndst, NULL, NULL);
     }
 
     xmlNodePtr n = ndst->children;
@@ -3570,6 +3675,8 @@ static gint copy_command(assuan_context_t ctx, gchar *line)
 	goto fail;
     }
 
+    rc = update_element_mtime(xmlDocGetRootElement(client->doc) == ndst->parent ? ndst : ndst->parent);
+
 fail:
     if (req)
 	g_strfreev(req);
@@ -3690,12 +3797,15 @@ static gint move_command(assuan_context_t ctx, gchar *line)
     if (!ndst)
 	goto fail;
 
+    update_element_mtime(nsrc->parent);
     xmlUnlinkNode(nsrc);
     ndst = xmlAddChildList(ndst, nsrc);
 
     if (!ndst)
 	rc = GPG_ERR_ENOMEM;
 
+    update_element_mtime(ndst->parent);
+
 fail:
     if (req)
 	g_strfreev(req);
@@ -3813,6 +3923,11 @@ gint command_startup(assuan_context_t ctx, const gchar *name)
 	    !g_ascii_strcasecmp(name, "UNSET"))
 	return 0;
 
+#ifdef WITH_PINENTRY
+    if (!(cl->opts & OPT_PINENTRY))
+	reset_pin_defaults(cl->pinentry);
+#endif
+
     cl->last_rc = rc = file_modified(cl);
 
     if (rc) {
