@@ -191,10 +191,12 @@ struct crypto_s {
     gsize keysize;
 };
 
-/* These are flags that are set by a client via the OPTION command. */
+/* These are flags that are set by a client option via the SET command. */
 #define OPT_ITERATIONS	0x1
 #define OPT_PINENTRY	0x2
 #define OPT_PINENTRY_TO	0x4
+#define OPT_CIPHER	0x8
+#define OPT_LOCK	0x10
 
 struct client_s {
     assuan_context_t ctx;
@@ -218,7 +220,7 @@ struct client_s {
     struct crypto_s *crypto;
     guchar opts;
     gpg_error_t last_rc;
-    gboolean lockonopen;
+    gboolean lock_on_open;
     gboolean rc_on_locked;
 };
 
@@ -229,18 +231,6 @@ pth_mutex_t rcfile_mutex;
 pth_mutex_t cn_mutex;
 GSList *cn_thread_list;
 
-#define log_write0 log_write
-
-#define log_write1(...) { \
-    if (get_key_file_integer("global", "log_level") >= 1) \
-	log_write(__VA_ARGS__); \
-    }
-
-#define log_write2(...) { \
-    if (get_key_file_integer("global", "log_level") >= 2) \
-	log_write(__VA_ARGS__); \
-    }
-
 void log_write(const gchar *fmt, ...);
 gpg_error_t send_error(assuan_context_t ctx, gpg_error_t pwmd_errno);
 gpg_error_t send_syserror(assuan_context_t ctx, int e);
@@ -267,6 +257,7 @@ void cleanup_unlink_cb(void *arg);
 void cleanup_attr_cb(void *arg);
 void cleanup_cancel_cb(void *arg);
 guint pwmd_cipher_str_to_cipher(const gchar *str);
+const gchar *pwmd_cipher_to_str(guint64 flags);
 file_header_internal_t *read_file_header(const gchar *filename, gboolean v1,
 	gpg_error_t *rc);
 
