@@ -19,6 +19,34 @@
 #ifndef MISC_H
 #define MISC_H
 
+gint get_key_file_integer(const gchar *section, const gchar *what);
+
+#define log_write0 log_write
+
+#define log_write1(...) { \
+    if (get_key_file_integer("global", "log_level") >= 1) \
+	log_write(__VA_ARGS__); \
+    }
+
+#define log_write2(...) { \
+    if (get_key_file_integer("global", "log_level") >= 2) \
+	log_write(__VA_ARGS__); \
+    }
+
+
+typedef enum {
+    OPT_NOARG,
+    OPT_ARG,
+    OPT_OPTARG,
+} option_type_t;
+
+struct argv_s {
+    const gchar *opt;
+    option_type_t type;
+    gpg_error_t (*func)(gpointer data, gpointer value);
+};
+
+gpg_error_t parse_options(gchar **line, struct argv_s *args[], gpointer data);
 gboolean strv_printf(gchar ***array, const gchar *fmt, ...);
 gchar **strvcatv(gchar **dst, gchar **src);
 gboolean valid_filename(const gchar *filename);
