@@ -485,30 +485,11 @@ static void *timeout_thread(void *arg)
     return NULL;
 }
 
-static void child_prepare(void *arg)
-{
-    gint i;
-    gint n = g_slist_length(cn_thread_list);
-    struct pinentry_s *pin = arg;
-
-    for (i = 0; i < n; i++) {
-	struct client_thread_s *cn = g_slist_nth_data(cn_thread_list, i);
-	pth_mutex_t *m;
-
-	cache_get_mutex(cn->cl->md5file, &m);
-	pth_mutex_release(m);
-    }
-
-    pth_mutex_release(&pin->status_mutex);
-    pth_mutex_release(&pin->cond_mutex);
-    free_client_list();
-}
-
 gpg_error_t pinentry_fork(assuan_context_t ctx)
 {
     struct client_s *client = assuan_get_pointer(ctx);
     struct pinentry_s *pin = client->pinentry;
-    gpg_error_t rc = 0;
+    gpg_error_t rc;
     gint p[2];
     pid_t pid;
     pinentry_key_s pk;
@@ -517,8 +498,7 @@ gpg_error_t pinentry_fork(assuan_context_t ctx)
     if (pipe(p) == -1)
 	return gpg_error_from_syserror();
 
-    pin->timeout = get_key_file_integer(pin->filename, "pinentry_timeout");
-    pth_atfork_push(NULL, NULL, child_prepare, pin);
+    reset_pin_defaults(pin);
     pid = pth_fork();
 
     switch (pid) {
@@ -526,9 +506,10 @@ gpg_error_t pinentry_fork(assuan_context_t ctx)
 	    rc = gpg_error_from_syserror();
 	    close(p[0]);
 	    close(p[1]);
-	    break;
+	    return rc;
 	case 0:
 	    close(p[0]);
+
 	    if (pin->timeout > 0 && pin->which == PINENTRY_OPEN) {
 		/*
 		 * Send the pid of the pinentry process back to pwmd so it can
@@ -537,16 +518,21 @@ gpg_error_t pinentry_fork(assuan_context_t ctx)
 		pk.status = PINENTRY_PID;
 		pk.error = pinentry_command(pin, &result, "GETINFO pid");
 
-		if (write_result(p[1], &pk, result))
+		if (write_result(p[1], &pk, result)) {
+		    free_client_list();
 		    _exit(EXIT_FAILURE);
+		}
 	    }
 
 	    pk.status = PINENTRY_RUNNING;
 	    pk.error = pinentry_getpin(pin, &result);
 
-	    if (write_result(p[1], &pk, result))
+	    if (write_result(p[1], &pk, result)) {
+		free_client_list();
 		_exit(EXIT_FAILURE);
+	    }
 
+	    free_client_list();
 	    _exit(EXIT_SUCCESS);
 	default:
 	    close(p[1]);
@@ -561,8 +547,7 @@ gpg_error_t pinentry_fork(assuan_context_t ctx)
      * the callback function which is called after the key has been read()
      * in pinentry_iterate().
      */
-    pth_atfork_pop();
-    return rc;
+    return 0;
 }
 
 gpg_error_t lock_pin_mutex(struct client_s *client)
@@ -650,7 +635,7 @@ void cleanup_pinentry(struct pinentry_s *pin)
 void reset_pin_defaults(struct pinentry_s *pin)
 {
     pin->enable = -1;
-    pin->timeout = get_key_file_integer("global", "pinentry_timeout");
+    pin->timeout = get_key_file_integer(pin->filename, "pinentry_timeout");
 }
 
 void set_pinentry_defaults(struct pinentry_s *pin)
@@ -817,10 +802,10 @@ int pinentry_iterate(struct client_s *cl, gboolean read_ready)
 		log_write(N_("read(): short byte count"));
 
 	    if (pk.error) {
-		if (cl->pinentry->tid) {
-		    cl->pinentry->status = PINENTRY_NONE;
+		cl->pinentry->status = PINENTRY_NONE;
+
+		if (cl->pinentry->tid)
 		    pth_cond_notify(&cl->pinentry->cond, FALSE);
-		}
 
 		if (cl->pinentry->which == PINENTRY_OPEN)
 		    cleanup_client(cl);
@@ -828,7 +813,6 @@ int pinentry_iterate(struct client_s *cl, gboolean read_ready)
 
 	    if (pk.error || pk.status == PINENTRY_RUNNING) {
 		reset(cl->pinentry);
-		unlock_file_mutex(cl);
 		unlock_pin_mutex(cl->pinentry);
 	    }
 
