@@ -1579,6 +1579,40 @@ guint pwmd_cipher_str_to_cipher(const gchar *str)
     return flags;
 }
 
+const gchar *pwmd_cipher_to_str(guint64 flags)
+{
+    if (flags & PWMD_CIPHER_AES128)
+	return "aes128";
+    else if (flags & PWMD_CIPHER_AES192)
+	return "aes192";
+    else if (flags & PWMD_CIPHER_AES256)
+	return "aes256";
+    else if (flags & PWMD_CIPHER_SERPENT128)
+	return "serpent128";
+    else if (flags & PWMD_CIPHER_SERPENT192)
+	return "serpent192";
+    else if (flags & PWMD_CIPHER_SERPENT256)
+	return "serpent256";
+    else if (flags & PWMD_CIPHER_CAMELLIA128)
+	return "camellia128";
+    else if (flags & PWMD_CIPHER_CAMELLIA192)
+	return "camellia192";
+    else if (flags & PWMD_CIPHER_CAMELLIA256)
+	return "camellia256";
+    else if (flags & PWMD_CIPHER_BLOWFISH)
+	return "blowfish";
+    else if (flags & PWMD_CIPHER_CAST5)
+	return "cast5";
+    else if (flags & PWMD_CIPHER_3DES)
+	return "3des";
+    else if (flags & PWMD_CIPHER_TWOFISH)
+	return "twofish256";
+    else if (flags & PWMD_CIPHER_TWOFISH128)
+	return "twofish128";
+
+    return NULL;
+}
+
 /* To be called after read_file_header(). This sets the wanted algorithm from
  * .flags */
 gpg_error_t init_client_crypto2(const char *filename,
