@@ -75,7 +75,6 @@ gboolean valid_xml_element(xmlChar *element)
 gpg_error_t new_root_element(xmlDocPtr doc, gchar *name)
 {
     xmlNodePtr root = xmlDocGetRootElement(doc);
-    xmlAttrPtr a;
     xmlNodePtr n;
     gchar *p = name;
 
@@ -90,8 +89,7 @@ gpg_error_t new_root_element(xmlDocPtr doc, gchar *name)
 
     n = xmlNewNode(NULL, (xmlChar *)"element");
     n = xmlAddChild(root, n);
-    a = xmlNewProp(n, (xmlChar *)"_name", (xmlChar *)p);
-    return 0;
+    return add_attribute(n, "_name", p);
 }
 
 xmlDocPtr create_dtd()
@@ -150,6 +148,21 @@ xmlChar *new_document()
     return xml;
 }
 
+xmlNodePtr find_element_node(xmlNodePtr node)
+{
+    xmlNodePtr n = node;
+
+    if (n && n->type == XML_ELEMENT_NODE)
+	return n;
+
+    for (n = node; n; n = n->next) {
+	if (n->type == XML_ELEMENT_NODE)
+	    return n;
+    }
+
+    return NULL;
+}
+
 /*
  * Lists root element names; the value of the attribute "_name" of an element
  * "element". If there's a target attribute both literal and non-literal
@@ -157,7 +170,8 @@ xmlChar *new_document()
  * cannot be used. There wouldn't be a way to get the literal an non-literal
  * element paths.
  */
-gpg_error_t list_root_elements(xmlDocPtr doc, GString **result)
+gpg_error_t list_root_elements(xmlDocPtr doc, GString **result,
+	gboolean verbose)
 {
     xmlNodePtr n = NULL;
     GSList *list = NULL;
@@ -191,7 +205,7 @@ gpg_error_t list_root_elements(xmlDocPtr doc, GString **result)
 	    goto fail;
 	}
 
-	tmp = g_strdup_printf("!%s", (gchar *)val);
+	tmp = g_strdup_printf("!%s%s", (gchar *)val, verbose ? find_element_node(n->children) ? " 1" : " 0" : "");
 
 	if (!tmp) {
 	    xmlFree(val);
@@ -211,7 +225,23 @@ gpg_error_t list_root_elements(xmlDocPtr doc, GString **result)
 	target = node_has_attribute(n, (xmlChar *)"target");
 
 	if (target) {
-	    gchar *t = g_strdup((gchar *)val);
+	    gchar *t;
+
+	    if (verbose) {
+		gchar **req = split_input_line((gchar *)target, "\t", 0);
+		xmlNodePtr tnode = find_root_element(doc, &req, &rc, NULL, 0, FALSE);
+
+		if (tnode && req[1]) {
+		    tnode = find_elements(doc, tnode->children, req+1, &rc,
+			    NULL, NULL, NULL, FALSE, 0, NULL, FALSE);
+		}
+
+		g_strfreev(req);
+		t = g_strdup_printf("%s%s", (gchar *)val,
+			tnode && find_element_node(tnode->children) ? " 1" : " 0");
+	    }
+	    else
+		t = g_strdup((gchar *)val);
 
 	    if (!t) {
 		xmlFree(val);
@@ -799,7 +829,13 @@ static gpg_error_t path_list_recurse(xmlDocPtr doc, xmlNodePtr node,
 	if (n->type != XML_ELEMENT_NODE)
 	    goto children;
 
-	if (strv_printf(&elements->elements, "%s\t!%s", elements->prefix, a) == FALSE) {
+	if (elements->verbose) {
+	    if (strv_printf(&elements->elements, "%s\t!%s%s", elements->prefix, a, find_element_node(n->children) ? " 1" : " 0") == FALSE) {
+		xmlFree(a);
+		return gpg_err_code_from_errno(ENOMEM);
+	    }
+	}
+	else if (strv_printf(&elements->elements, "%s\t!%s", elements->prefix, a) == FALSE) {
 	    xmlFree(a);
 	    return gpg_err_code_from_errno(ENOMEM);
 	}
@@ -820,6 +856,7 @@ static gpg_error_t path_list_recurse(xmlDocPtr doc, xmlNodePtr node,
 
 	    if (max_recursion_depth >= 1 && elements->depth > max_recursion_depth) {
 		xmlChar *t = xmlGetNodePath(n);
+
 		log_write("%s: %s", pwmd_strerror(GPG_ERR_ELOOP), t);
 		xmlFree(t);
 		xmlFree(target);
@@ -827,7 +864,33 @@ static gpg_error_t path_list_recurse(xmlDocPtr doc, xmlNodePtr node,
 		return GPG_ERR_ELOOP;
 	    }
 
-	    if (strv_printf(&elements->elements, "%s\t%s", elements->prefix, a) == FALSE) {
+	    if (elements->verbose) {
+		gchar **req = split_input_line((gchar *)target, "\t", 0);
+		xmlNodePtr tnode = find_root_element(doc, &req, &rc, NULL, 0, FALSE);
+
+		if (!tnode) {
+		    g_strfreev(req);
+		    xmlFree(a);
+		    xmlFree(target);
+		    return rc;
+		}
+
+		tnode = find_elements(doc, tnode->children, req+1, &rc, NULL, NULL, NULL, FALSE, 0, NULL, FALSE);
+		g_strfreev(req);
+
+		if (!tnode) {
+		    xmlFree(a);
+		    xmlFree(target);
+		    return rc;
+		}
+
+		if (strv_printf(&elements->elements, "%s\t%s%s", elements->prefix, a, find_element_node(tnode->children) ? " 1" : " 0") == FALSE) {
+		    xmlFree(a);
+		    xmlFree(target);
+		    return gpg_err_code_from_errno(ENOMEM);
+		}
+	    }
+	    else if (strv_printf(&elements->elements, "%s\t%s", elements->prefix, a) == FALSE) {
 		xmlFree(a);
 		xmlFree(target);
 		return gpg_err_code_from_errno(ENOMEM);
@@ -842,28 +905,34 @@ static gpg_error_t path_list_recurse(xmlDocPtr doc, xmlNodePtr node,
 	    }
 
 	    if (update_element_list(elements) == FALSE) {
+		g_free(tmp);
 		xmlFree(a);
 		xmlFree(target);
 		return gpg_err_code_from_errno(ENOMEM);
 	    }
 
-	    elements->prefix = tmp;
-	    elements->resolving = TRUE;
-	    rc = create_path_list(doc, elements, (gchar *)target);
-	    xmlFree(target);
-	    elements->resolving = r;
-	    elements->depth--;
-	    g_free(tmp);
-	    elements->prefix = save;
+	    if (elements->recurse) {
+		if (elements->verbose)
+		    tmp[strlen(tmp)-2] = 0;
 
-	    if (rc) {
-		xmlFree(a);
-		return rc;
+		elements->prefix = tmp;
+		elements->resolving = TRUE;
+		rc = create_path_list(doc, elements, (gchar *)target);
+		xmlFree(target);
+		elements->resolving = r;
+		elements->depth--;
+		g_free(tmp);
+		elements->prefix = save;
+
+		if (rc) {
+		    xmlFree(a);
+		    return rc;
+		}
 	    }
 	}
 
 children:
-	if (n->children) {
+	if (n->children && elements->recurse) {
 	    gchar *tmp = g_strdup_printf("%s\t!%s", elements->prefix, a);
 	    gchar *save = elements->prefix;
 
@@ -977,7 +1046,13 @@ gpg_error_t create_path_list(xmlDocPtr doc, struct element_list_s *elements,
 	    goto fail;
 	}
 
-	if (strv_printf(&elements->elements, "%s", elements->prefix) == FALSE) {
+	if (elements->verbose) {
+	    if (strv_printf(&elements->elements, "%s%s", elements->prefix, find_element_node(n->children) ? " 1" : " 0") == FALSE) {
+		rc = gpg_err_code_from_errno(ENOMEM);
+		goto fail;
+	    }
+	}
+	else if (strv_printf(&elements->elements, "%s", elements->prefix) == FALSE) {
 	    rc = gpg_err_code_from_errno(ENOMEM);
 	    goto fail;
 	}
@@ -1276,3 +1351,17 @@ gpg_error_t update_element_mtime(xmlNodePtr n)
 {
     return add_attribute(n, NULL, NULL);
 }
+
+gpg_error_t unlink_node(xmlNodePtr n)
+{
+    gpg_error_t rc = 0;
+
+    if (!n)
+	return rc;
+
+    if (n->parent)
+	rc = update_element_mtime(n->parent);
+
+    xmlUnlinkNode(n);
+    return rc;
+}
